#!/usr/bin/perl -w

# Copyright 2016, 2017 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  See the file COPYING.  If not, see
# <http://www.gnu.org/licenses/>.

use strict;
use 5.006;
use Carp 'croak';
use List::Util 'min','max';
use Math::BaseCnv 'cnv';
use Math::PlanePath::CCurve;
use Math::NumSeq::PlanePathDelta;
use Tie::IxHash;
use Test::More;

use FindBin;
use lib $FindBin::Bin;
use MyCCurve ':all';

plan tests => 92;


#------------------------------------------------------------------------------
# left boundary

{
  my @L_samples_generated = (1,2,4,8,16,32,64,124,202,308,450,648,918,1300,1826,2576,3614,5100,7162,10120,14230,20132,28338,40128,56526);
  is_deeply([map {L($_)} 0 .. $#L_samples_generated],
            \@L_samples_generated);
}


{
  my @states = NonLpred_k_flat()->MyFLAT::get_eventually_accepting();
  is_deeply([sort {$a<=>$b} @states],
            [ 17,21, 26,29, 32,33, 36,37,
              45,46,47,48 # d
            ]);
}

{
  # L vs Lpred_k_flat()
  # Lpred_k_flat() counts also 1 each end of each even part k>=6 

  my $max_k = 30;
  is_deeply([MyFLAT::FLAT_count_contains (Lpred_k_flat(), $max_k)],
            [map {L($_) + 2*max(0,$_-7)} 0 .. $max_k]);


  # $,=' ';
  # print MyFLAT::FLAT_count_contains(Lpred_k_flat(), $max_k);
  # print "\n";
  # print map {L($_) + 2*max(0,$_-7)} 0 .. $max_k;
  # print "\n";
}

sub Lpred_k_hash_by_traverse {
  my ($k, $bothways) = @_;
  my $path = Math::PlanePath::CCurve->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);

  my $x = my $y = 0;
  my $dx = 0;
  my $dy = -1;
  # tie my %seen, 'Tie::IxHash';
  my %seen;
 LOCATION: until ($x == $x_hi && $y == $y_hi) {
    ($dx,$dy) = xy_rotate_plus90($dx,$dy);  # try to follow to the left
    foreach (1 .. 4) {
      my $x2 = $x+$dx;
      my $y2 = $y+$dy;
      my @n_list = $path->xyxy_to_n_list_either($x,$y, $x2,$y2);
      @n_list = grep {$_ < $n_hi} @n_list;
      if (@n_list) {
        if (! $bothways) {
          @n_list = $path->xyxy_to_n_list($x,$y, $x2,$y2)
            or @n_list = $path->xyxy_to_n_list($x2,$y2, $x,$y);
        }
        @seen{@n_list} = (); # hash slice
        $x = $x2;
        $y = $y2;
        next LOCATION;
      }
      ($dx,$dy) = xy_rotate_minus90($dx,$dy);
    }
  }
  return \%seen;
}

foreach my $k (0 .. 10) {
  my $L = L($k);
  my $by_traverse_oneway = Lpred_k_hash_by_traverse($k,0);
  my @by_traverse_oneway = keys %$by_traverse_oneway;
  is (scalar(@by_traverse_oneway), $L, "L vs traverse k=$k");
}

foreach my $k (0 .. 10) {
  my $by_traverse_bothways = Lpred_k_hash_by_traverse($k,1);

  my $path = Math::PlanePath::CCurve->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my $f = Lpred_k_flat();
  my $diff = 0;
  foreach my $n ($n_lo .. $n_hi-1) {
    my $by_t = exists $by_traverse_bothways->{$n} ? 1 : 0;
    my $n2 = sprintf '%0*b', $k,$n;
    my $by_f = $f->contains($n2) ? 1 : 0;
    if ($by_t != $by_f) {
      diag("k=$k n=$n [$n2]  by_t=$by_t by_f=$by_f");
      last if ++$diff > 10;
    }
  }
  is ($diff, 0, "Lpred_k_flat() vs traverse k=$k");
}

sub Lpred_k_list_by_flat {
  my ($k) = @_;
  my $f = Lpred_k_flat();
  my @list = ('');
  foreach (1 .. $k) {
    my @new_list;
    foreach my $str (@list) {
      foreach my $bit ('0','1') {
        my $new_str = $str . $bit;
        if ($f->contains($new_str)) {
          push @new_list, $new_str;
        }
      }
    }
    @list = @new_list;
  }
  return map {oct('0b'.$_)} @list;
}

foreach my $k (0 .. 12) {
  my $by_traverse = Lpred_k_hash_by_traverse($k,1);
  my @by_traverse = keys %$by_traverse;
  @by_traverse = sort {$a<=>$b} @by_traverse;
  my @by_flat = Lpred_k_list_by_flat($k);
  is_deeply(\@by_flat, \@by_traverse, "Lpred_k_flat() list vs traverse k=$k");

  # print join(', ',@by_traverse),"\n";
  # print join(', ',@by_flat),"\n";
}

#------------------------------------------------------------------------------
# Rpred_k_flat()

ok(Rpred_k_flat()->equals( NonRpred_k_flat()->complement ));
ok(RpredInf_flat()->equals( NonRpredInf_flat()->complement ));

{
  # NonRpred is 00 to 00 odd distance apart or 11 to 11 odd distance apart

  my $odd_dist = '(0|1) ( (0|1) (0|1) )*';
  my $zeros_odd_dist = "00 $odd_dist 00  | 000";
  my $ones_odd_dist  = "11 $odd_dist 11  | 111";

  {
    my $f = FLAT::Regex->new($odd_dist);
    ok(  $f->contains('0'));
    ok(! $f->contains('11'));
    ok(  $f->contains('111'));
  }
  {
    my $f = FLAT::Regex->new($ones_odd_dist);
    ok(  $f->contains('111'));
    ok(  $f->contains('11011'));
    MyFLAT::FLAT_show_breadth($f->as_dfa, 4);
  }

  # NFA since union of DFAs not quite right in FLAT 0.9.1
  require MyFLAT;
  my $f = FLAT::Regex->new(" 0 (0|1)* ($ones_odd_dist) (0|1)*")
    ->as_nfa
    ->MyFLAT::minimize(verbose=>1);

  $f = $f->union(FLAT::Regex->new(" 11*   (0|1)* ($zeros_odd_dist) (0|1)*")
                 ->as_nfa
                 ->MyFLAT::minimize(verbose=>1))
    ->as_nfa
    ->MyFLAT::minimize(verbose=>1);

  $f = $f->union(FLAT::Regex->new(" 00* 1 (0|1)* ($zeros_odd_dist) (0|1)*")
                 ->as_nfa
                 ->MyFLAT::minimize(verbose=>1))
    ->as_nfa
    ->MyFLAT::minimize(verbose=>1);

  $f = $f->union(FLAT::Regex->new(" 11* 0 (0|1)* ($ones_odd_dist) (0|1)*")
                 ->as_nfa
                 ->MyFLAT::minimize(verbose=>1))
    ->as_nfa
    ->MyFLAT::minimize(verbose=>1);

  ok($f->equals(NonRpred_k_flat()));
  MyFLAT::FLAT_check_is_equal($f, NonRpred_k_flat());
}


#------------------------------------------------------------------------------


{
  my @R_samples = (1, 2, 4, 8, 14, 24, 38, 60, 90, 136, 198, 292, 418);
  foreach my $k (0 .. $#R_samples) {
    is (R($k), $R_samples[$k]);
    # or die "k=$k  got ",R($k)," want ",$R_samples[$k];
  }
}

{
  is (PartN(4), 5);
  my @PartN_samples = (0,1,1,3,5,11,21,43);  # A001045
  foreach my $k (0 .. $#PartN_samples) {
    is (PartN($k), $PartN_samples[$k]);
    # or die "k=$k  got ",PartN($k)," want ",$PartN_samples[$k];
  }
}

{
  my @TopMiddlePre_samples = (0, 0, 1, 3, 6, 13, 26, 53, 106, 213, 426);
  foreach my $k (0 .. $#TopMiddlePre_samples) {
    is (TopMiddlePre($k), $TopMiddlePre_samples[$k]);
    # or die "k=$k  got ",TopMiddlePre($k)," want ",$TopMiddlePre_samples[$k];
  }
}
{
  my @TopMiddlePost_samples = (0, 1, 2, 4, 9, 18, 37, 74, 149, 298, 597);
  foreach my $k (0 .. $#TopMiddlePost_samples) {
    is (TopMiddlePost($k), $TopMiddlePost_samples[$k]);
    # or die "k=$k  got ",TopMiddlePost($k)," want ",$TopMiddlePost_samples[$k];
  }
}


#------------------------------------------------------------------------------
# turn()

{
  my $path = Math::PlanePath::CCurve->new;
  my $seq = Math::NumSeq::PlanePathDelta->new(planepath_object=>$path,
                                              delta_type => 'Dir4');
  my $bad = 0;
  my $prev_dir;
  foreach (1 .. 1000) {
    my ($n,$dir_by_seq) = $seq->next;
    my $dir_by_func = dir($n);
    my $turn_by_func = ($n==0 ? -999 : turn($n));
    my $turn_by_seq = (defined $prev_dir ? $dir_by_seq - $prev_dir : -999);
    unless ($dir_by_seq == ($dir_by_func % 4)) {
      diag "n=$n dir()=$dir_by_func seq=$dir_by_seq";
      last if $bad++ > 10;
    }
    unless (($turn_by_seq % 4) == ($turn_by_func % 4)) {
      diag "n=$n turn()=$turn_by_func seq=$turn_by_seq";
      last if $bad++ > 10;
    }
    $prev_dir = $dir_by_seq;
  }
  is($bad, 0);
}

#------------------------------------------------------------------------------
# turn_at_xy()

{
  my $path = Math::PlanePath::CCurve->new;
  my $bad = 0;
  foreach my $n (1 .. 1000) {
    my $turn_at_n = turn($n);
    my ($x,$y) = $path->n_to_xy($n);
    my $turn_at_xy = turn_at_xy($x,$y);
    unless ($turn_at_n == $turn_at_xy) {
      diag "n=$n turn()=$turn_at_n turn_at_xy()=$turn_at_xy";
      last if $bad++ > 10;
    }
  }
  is($bad, 0);
}


#------------------------------------------------------------------------------
exit 0;
