#!/usr/bin/perl -w

# Copyright 2014, 2015, 2019 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


use 5.004;
use strict;
use FindBin;
use List::Util 'sum';
use Math::PlanePath::CCurve 117; # v.117 for level_to_n_range()

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyCCurve 'xy_rotate_minus90';
use MyPrintwrap ':all';

# uncomment this to run the ### lines
# use Smart::Comments;


my $max_k = 16;
if (@ARGV) { $max_k = shift @ARGV; }

print "% generated by tools/$FindBin::Script $max_k\n";
my $path = Math::PlanePath::CCurve->new;

# sides in square regions count

#   4  |   1    |  4          3  |   2    |  3
# ----- <------- -----      ----- -------> -----
#   3  |   2    |  3          4  |   1    |  4
#
# real((x+I*y)*conj(xh+yh*I))
# imag((x+I*y)*conj(xh+yh*I))

my @QR_values;
my @TR_values;
foreach my $k (0 .. $max_k) {
  my ($n_lo, $n_hi) = $path->level_to_n_range($k);
  my ($x_hi,$y_hi) = $path->n_to_xy($n_hi);
  ### $n_hi

  my $segment_to_QR = sub {
    my ($x,$y) = @_;
    ### segment_to_QR(): "$x, $y"
    die if $x == 0 || $x == $n_hi || $y == 0;

    ($x, $y) = ($x*$x_hi + $y*$y_hi, # complex multiply x,y * conj(x_hi,y_hi)
                - $x*$y_hi + $y*$x_hi);
    ### multiply: "$x, $y"

    if ($y < 0) {
      return ($x > 0 && $x < $n_hi ? 1 : 4);
    } else {
      return ($x > 0 && $x < $n_hi ? 2 : 3);
    }
  };

  my $segment_to_TR = sub {
    my ($x,$y) = @_;
    ### segment_to_QR(): "$x, $y"
    die if $x == 0 || $x == $n_hi || $y == 0;

    ($x, $y) = ($x*$x_hi + $y*$y_hi, # complex multiply x,y * conj(x_hi,y_hi)
                - $x*$y_hi + $y*$x_hi);
    ### multiply: "$x, $y"
    if ($x > $n_hi) {
      $x = $n_hi - $x;
    }
    if ($x < 0) {
      if ($y < 0) {
        return ($y < $x ? 7 : 6);
      } else {
        return (-$x > $y ? 5 : 4);
      }
    }

    if ($y > 0) {
      if ($x > $y && $n_hi-$x > $y) { return 2; }
      return 3;
    }
    if ($y < 0) {
      $y = - $y;
      if ($x > $y && $n_hi-$x > $y) { return 1; }
      $y = $n_hi - $y;
      if ($x > $y && $n_hi-$x > $y) { return 9; }
      return 8;
    }
  };

  my @QR_counts = ((0) x 5);
  my @TR_counts = ((0) x 10);
  my $x = 0;
  my $y = 0;
  foreach my $n ($n_lo .. $n_hi-1) {
    my ($dx,$dy) = $path->n_to_dxdy($n);
    {
      my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
      my $qr = $segment_to_QR->($x+$dx/2+$rx/4, $y+$dy/2+$ry/4);
      $QR_counts[$qr]++;
    }
    {
      my ($rx,$ry) = xy_rotate_minus90($dx,$dy);
      my $tr = $segment_to_TR->($x+$dx/2+$rx/4, $y+$dy/2+$ry/4);
      $TR_counts[$tr]++;
    }
    $x += $dx;
    $y += $dy;
  }

  {
    my $sum = sum(@QR_counts);
    die if $sum != 2**$k;
  }
  {
    my $sum = sum(@TR_counts);
    die if $sum != 2**$k;
  }

  $QR_counts[3] /= 2;
  $QR_counts[4] /= 2;

  $TR_counts[3] /= 2;
  $TR_counts[4] /= 2;
  $TR_counts[5] /= 2;
  $TR_counts[6] /= 2;
  $TR_counts[7] /= 2;
  $TR_counts[8] /= 2;

  push @QR_values, \@QR_counts;
  push @TR_values, \@TR_counts;
}

foreach my $s (1 .. 4) {
  print "% GP-DEFINE  QR${s}_generated_samples = [";
  my $join = '';
  foreach my $k (0 .. $#QR_values) {
    print $join,$QR_values[$k]->[$s];
    $join = ',';
  }
  print "];\n";
}

foreach my $s (1 .. 9) {
  print "% GP-DEFINE  TR${s}_generated_samples = [";
  my $join = '';
  foreach my $k (0 .. $#TR_values) {
    print $join,$TR_values[$k]->[$s];
    $join = ',';
  }
  print "];\n";
}

print "% end generated\n";
exit 0;
