#!/usr/bin/perl -w

# Copyright 2014, 2015, 2018 Kevin Ryde
#
# This file is free software; you can redistribute it and/or modify it
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 3, or (at your option) any later
# version.
#
# This file is distributed in the hope that it will be useful, but
# WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY
# or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License
# for more details.
#
# You should have received a copy of the GNU General Public License along
# with this file.  If not, see <http://www.gnu.org/licenses/>.


# Create sample picture image.png.

use 5.004;
use strict;
use FindBin;
use File::chdir;
use File::stat;
use File::Temp;
use IPC::Run;
use List::Util 'max','min';

use Math::PlanePath::CCurve 117; # v.117 for level_to_n_range()

use FindBin;
BEGIN { unshift @INC, $FindBin::Bin }
use MyCCurve 'xy_rotate_minus90';

# uncomment this to run the ### lines
# use Smart::Comments;


my $level = 12;
my $scale_factor = 3.5;
my $scale_pixels = 2;
my $chamfer = 0;
my $AlphaBits = 4;   # 1,2,4

my $program_filename = File::Spec->catfile($FindBin::Bin,
                                           $FindBin::Script);
my $png_filename = File::Spec->catfile($FindBin::Bin,
                                       File::Spec->updir,
                                       'image.png');

{
  my $background = 'white';
  my $foreground = 'black';
  my $x_margin = 10;
  my $y_margin = 10;

  my $path = Math::PlanePath::CCurve->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);
  print "N= $n_lo to $n_hi\n";
  my (@x,@y);
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);
    foreach (-1 .. int($level/2)) { ($x,$y) = xy_rotate_minus90($x,$y); }
    $x[$n] = $x * $scale_pixels;
    $y[$n] = $y * $scale_pixels;
  }
  my $x_min = min(@x);
  my $y_min = min(@y);
  my $x_max = max(@x);
  my $y_max = max(@y);
  ### $x_min
  ### $y_min
  ### $x_max
  ### $y_max

  my $width  = ($x_max - $x_min) + 1 + 2*$x_margin;
  my $height = ($y_max - $y_min) + 1 + 2*$y_margin;
  my $x_base = -$x_min + $x_margin;
  my $y_base = -$y_min + $y_margin;
  print "X $x_min to $x_max   Y $y_min to $y_max  width=$width height=$height  X base $x_base Y base $y_base\n";


  require Image::Base::GD;
  my $image = Image::Base::GD->new (-height => $height,
                                    -width  => $width);
  $image->rectangle(0,0, $width-1,$height-1, $background);
  foreach my $n ($n_lo .. $n_hi-1) {
    my $x1 = $x[$n];
    my $y1 = $y[$n];
    my $x2 = $x[$n+1];
    my $y2 = $y[$n+1];
    $x1 += $x_base;
    $y1 += $y_base;
    $x2 += $x_base;
    $y2 += $y_base;
    $y1 = $height-1 - $y1;
    $y2 = $height-1 - $y2;
    ### line: "$x1,$y1 to $x2,$y2"
    $image->line($x1,$y1, $x2,$y2, $foreground);
  }
  $image->save($png_filename);
}

if (0) {
  my $temp = File::Temp->newdir;
  my $tempdir = $temp->dirname;
  local $CWD = $tempdir;

  open my $tex_fh, '>', 'image.tex' or die;

  my $rotate = -45*($level-4);
  my $scale = $scale_factor * 1/sqrt(2)**$level;
  print $tex_fh <<"HERE";
\\documentclass[a4paper]{article}
\\pagestyle{empty}
\\usepackage{tikz}
\\begin{document}
\\begin{tikzpicture}[scale=$scale,rotate=$rotate]
  \\draw
HERE

  my $path = Math::PlanePath::CCurve->new;
  my ($n_lo, $n_hi) = $path->level_to_n_range($level);
  my $join = '';
  foreach my $n ($n_lo .. $n_hi) {
    my ($x,$y) = $path->n_to_xy($n);

    if ($n != $n_lo && $n != $n_hi) {
      my ($dx,$dy) = $path->n_to_dxdy($n);
      my ($prev_dx,$prev_dy) = $path->n_to_dxdy($n-1);
      my $sx = $x - $chamfer*$prev_dx;
      my $sy = $y - $chamfer*$prev_dy;
      my $tx = $x + $chamfer*$dx;
      my $ty = $y + $chamfer*$dy;
      print $tex_fh "  $join($sx,$sy)--($tx,$ty)\n";
    } else {
      print $tex_fh "  $join($x,$y)\n";
    }
    $join = '--';
  }

  print $tex_fh <<'HERE';
;
\end{tikzpicture}
\end{document}
HERE
  close $tex_fh or die;
  IPC::Run::run(['cat','image.tex']);

  IPC::Run::run(['latexmk','-pdf','image.tex'], '<',File::Spec->devnull);

  IPC::Run::run(['pdfcrop',
                 '--verbose',
                 '--margins','10',
                 'image.pdf',
                 'image-cropped.pdf'],
                '<', File::Spec->devnull)
      or die;
  rename 'image-cropped.pdf', 'image.pdf' or die;

  print "gs\n";
  IPC::Run::run(['gs',
                 '-q', # quiet, cf -dQUIET
                 '-dNOPAUSE',
                 '-r100',
                 '-sDEVICE=pnggray',
                 "-dTextAlphaBits=$AlphaBits",     # grey scale edge blurring
                 "-dGraphicsAlphaBits=$AlphaBits",
                 "-sOutputFile=$png_filename",
                 'image.pdf'],
                '<', File::Spec->devnull)
      or die;
  print "\n";
}


# PNG spec 11.3.4.2 suggests RFC822 (or rather RFC1123) for CreationTime
use constant STRFTIME_FORMAT_RFC822 => '%a, %d %b %Y %H:%M:%S %z';
my $stat = File::stat::stat($program_filename)
  or die "Oops, cannot stat $program_filename: $!";
my $creation_time = POSIX::strftime(STRFTIME_FORMAT_RFC822,
                                    localtime($stat->mtime));

pngtextadd($png_filename, 'Author',    'Kevin Ryde');
pngtextadd($png_filename, 'Generator', "tools/$FindBin::Script");
pngtextadd($png_filename, 'Title',     'Iterations of the C Curve - sample image');
pngtextadd($png_filename, 'Creation Time', $creation_time);
pngtextadd($png_filename, 'Copyright', 'Copyright 2015, 2018 Kevin Ryde');
pngtextadd($png_filename, 'Homepage',  'http://user42.tuxfamily.org/c-curve/index.html');

print "uncompressed ",-s $png_filename,"\n";
IPC::Run::run(['optipng','-quiet','-o5',$png_filename]);
print "compressed   ",-s $png_filename,"\n";


if (($ARGV[0]||0) eq 'show') {
  # IPC::Run::run(['xpdf','image.pdf']);
  IPC::Run::run(['xzgv',$png_filename]);
}

exit 0;

# add text to the png image in $png_filename
sub pngtextadd {
  my ($filename, $keyword, $value) = @_;
  system('pngtextadd', "--keyword=$keyword", "--text=$value", $filename) == 0
    or die "system(pngtextadd)";
}
